#include <iostream>
#include <utility>
#include "bix.hpp"
#include <tuple>
#include <assert.h>
#include <conio.h>

int my_add(lua_State *l) {
    float x = (float)luaL_checknumber(l, 1); // reads a value from the stack,
                                     // verifying it is a number
    float y = (float)luaL_checknumber(l, 2);
    const float result = x + y;
    lua_pushnumber(l, result);
    return 1; // return the arity of this function
}

void lua_test()
{
	// open a lua context
	lua_State *l = luaL_newstate();

	// bind the function pointer to global "c_my_add"
	lua_register(l, "c_my_add", &my_add);
}


int foo(int x, int y, int z)
{
	return x + y + z;
}

using std::cout;
using std::endl;

template<int ...> struct seq {};

template<int N, int ...S> struct gens : gens<N - 1, N - 1, S...> {};

template<int ...S> struct gens<0, S...>{ typedef seq<S...> type; };

template <typename ...Args>
struct save_it_for_later
{
	std::tuple<Args...> params;
	int(*func)(Args...);

	double delayed_dispatch()
	{
		return callFunc(typename gens<sizeof...(Args)>::type());
	}

	template<int ...S>
	double callFunc(seq<S...>)
	{
		return func(std::get<S>(params) ...);
	}
};

template<typename ...args>
struct boo{
	static std::tuple<args...> make(args... some_args)
	{
		return std::make_tuple(some_args...);
	}
};

int main()
{
	std::cout<<"Bix\n";

	//gens<3> g;
	//gens<3>::type s;
	std::tuple<int, int, int> t = std::make_tuple(1, 2 , 3);
	save_it_for_later<int, int, int> saved = { t, foo };
	cout << "\n[["<< saved.delayed_dispatch() << "]]"<<endl;

	Bix::State state;
	luaL_openlibs(state._l);
	luaL_dostring(state._l, "print('xxx');");

	
	
	// Push function name to the stack
	lua_getglobal(state._l, "print");
	auto num_args = state.Push(50, "xass", 50.66, "dogonast", 23423432);
	lua_call(state._l, num_args, 0);
	std::cout << "\nnum_args = " << num_args;


	state.Push(11, "hi", true);

	int a;
	std::string b;
	bool c ;
	std::tie<int, std::string, bool>(a,b,c) = state.Pop<int,std::string,bool>();

	std::cout << "\nValues [" << a << "|" << b.c_str() << "|" << c << "]";
	assert(a == 11 && b == "hi" && c);
	


	state.Push(121);
	int d = state.Pop<int>();
	assert(d == 121);
	std::cout << "\nThings work ";

	std::cout << "\nCalling lua print\n";
	state.Call("print", 12, "this is text", std::string("via std::string"), (float)1.11, (double)2.22);

	luaL_dostring(state._l,"function add(a,b) print('-->lua add function called') return a+b end");
	int sum = state.Call<int>("add", 1, 2);
	std::cout << "Sum of 1+2 = " << sum;

	luaL_dostring(state._l, "function reflect2(a,b) print('-->lua reflect2 function called') return a,b end");
	int r_a;
	std::string r_b;
	std::tie<int, std::string>(r_a, r_b) = state.Call<int, std::string>("reflect2", 102, "bambam");
	assert(r_a == 102 && r_b == "bambam");

	luaL_dostring(state._l, "function outputx() print('this is some text') end");
	state.Call("outputx");


	auto bx = boo<int, int>::make(1, 2);

	_getch();

	return 0;
}